/* Copyright (C) 2025 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   KfpV0Finder.cxx
/// \date   01.02.2025
/// \brief  A V0 finding algorithm (implementation)
/// \author Sergei Zharko <s.zharko@gsi.de>

#include "kfp/KfpV0Finder.h"

#include "global/RecoResults.h"

#include <algorithm>
#include <limits>
#include <sstream>

using cbm::algo::RecoResults;
using cbm::algo::kfp::V0Finder;

// ---------------------------------------------------------------------------------------------------------------------
//
std::vector<double> V0Finder::CollectDca(const RecoResults& recoEvent) const
{
  std::vector<double> dca(recoEvent.trackStsHitIndices.size(), std::numeric_limits<double>::signaling_NaN());
  const auto& stsHitIndices = recoEvent.trackStsHitIndices;
  for (size_t iTrk = 0; iTrk < stsHitIndices.size(); ++iTrk) {
    const auto& stsHitIndicesInTrack = stsHitIndices[iTrk];
    if (stsHitIndicesInTrack.size() < 2) {  // less then two sts hits
      continue;
    }
    auto [iPtFst, iHitFst] = stsHitIndicesInTrack[0];
    auto [iPtSnd, iHitSnd] = stsHitIndicesInTrack[1];
    dca[iTrk]              = EstimateDca(recoEvent.stsHits[iPtFst][iHitFst], recoEvent.stsHits[iPtSnd][iHitSnd]);
  }
  return dca;
}

// ---------------------------------------------------------------------------------------------------------------------
//
std::vector<double> V0Finder::CollectT0(gsl::span<const bmon::Hit> bmonHits) const
{
  std::vector<double> t0;
  t0.reserve(bmonHits.size());
  std::transform(bmonHits.begin(), bmonHits.end(), std::back_inserter(t0), [&](const auto& h) { return h.GetTime(); });
  return t0;
}

// ---------------------------------------------------------------------------------------------------------------------
//
double V0Finder::EstimateDca(const sts::Hit& fst, const sts::Hit& snd) const
{
  double factor{(fst.Z() - fOrigin[2]) / (snd.Z() - fst.Z())};
  double dcaX{fst.X() - fOrigin[0] - factor * (snd.X() - fst.X())};
  double dcaY{fst.Y() - fOrigin[1] - factor * (snd.Y() - fst.Y())};
  return std::sqrt(dcaX * dcaX + dcaY * dcaY);
}

// ---------------------------------------------------------------------------------------------------------------------
//
void V0Finder::Init() {}

// ---------------------------------------------------------------------------------------------------------------------
//
CbmEventTriggers V0Finder::Process(const RecoResults& recoEvent)
{
  CbmEventTriggers res;
  fEventMonitor.Reset();
  fEventMonitor.StartTimer(ETimer::Event);
  fEventMonitor.IncrementCounter(ECounter::EventsTotal);

  // ----- Define T0
  // So far we cannot preselect a hit from multiple ones, we will be using all of them iteratively to find lambdas
  auto vT0 = CollectT0(recoEvent.bmonHits[fBmonPartitionIndex]);
  if (vT0.empty()) {
    fEventMonitor.IncrementCounter(ECounter::EventsWoTzero);
    return res;
  }

  //L_(info) << "------- Event: ";

  // ----- Estimate DCA of tracks
  // If a track has less then two STS hits, and undefined DCA value is stored
  std::vector<double> vDca = CollectDca(recoEvent);

  // ***** DEBUG: BEGIN
  //for (auto dca: vDca) {
  //  L_(info) << "dca=" << dca;
  //}
  // ***** DEBUG: END


  fEventMonitor.StopTimer(ETimer::Event);
  return res;
}
