/* Copyright (C) 2025 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   KfpV0Finder.h
/// \date   01.02.2025
/// \brief  A V0 finding algorithm
/// \author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "CbmEventTriggers.h"
#include "KFParticleTopoReconstructor.h"
#include "global/RecoResults.h"
#include "kfp/KfpV0FinderMonitor.h"

#include <limits>
#include <memory>


namespace cbm::algo::kfp
{
  /// \class V0Finder
  /// \brief A V0-finding algorithm
  class V0Finder {
   public:
    /// \brief Default constructor
    V0Finder() = default;

    /// \brief Copy constructor
    V0Finder(const V0Finder&) = delete;

    /// \brief Move constructor
    V0Finder(V0Finder&&) = delete;

    /// \brief Destructor
    ~V0Finder() = default;

    /// \brief Copy assignment operator
    V0Finder& operator=(const V0Finder&) = delete;

    /// \brief Move assignment operator
    V0Finder& operator=(V0Finder&&) = delete;

    /// \brief  Adds particle to reconstruction list
    /// \param  pdg  A PDG code of the particle to be reconstructed
    void AddDecayToReconstructionList(int pdg) { GetKFParticleFinder()->AddDecayToReconstructionList(pdg); }

    /// \brief  Mutable access to the KfParticleFinder of the run topology reconstructor
    KFParticleFinder* GetKFParticleFinder() { return fpTopoReconstructor->GetKFParticleFinder(); }

    /// \brief  Constant access to the KfParticleFinder of the run topology reconstructor
    const KFParticleFinder* GetKFParticleFinder() const { return fpTopoReconstructor->GetKFParticleFinder(); }

    /// \brief  Gets monitor data
    const V0FinderMonitorData_t& GetEventMonitor() const { return fEventMonitor; }

    /// \brief Initializes the instance (called in the beginning of the run)
    void Init();

    /// \brief  Processes a reconstructed data sample, returns a collection of fired triggers
    CbmEventTriggers Process(const RecoResults& recoEvent);

    /// \brief  Sets an address of a reference BMON diamond
    /// \param  iPartition  An index of the BMON hit partition: used to selected the hits from the particular config
    void SetBmonPartitionIndex(int iPartition) { fBmonPartitionIndex = iPartition; }

    /// \param  dca  DCA [cm]
    void SetMinPionDca(double dca) { fMinPionDca = dca; }

    /// \brief  Sets minimal proton DCA to primary vertex
    /// \param  dca  DCA [cm]
    void SetMinProtonDca(double dca) { fMinProtonDca = dca; }

    /// \brief  Sets origin
    /// \param  x  X-coordinate of the origin [cm]
    /// \param  y  Y-coordinate of the origin [cm]
    /// \param  z  Z-coordinate of the origin [cm]
    // FIXME: for now origin is defined as the target center, later it can be changed
    void SetOrigin(double x, double y, double z) { fOrigin = {x, y, z}; }

    /// \brief  Sets minimal pion DCA to primary vertex
    /// \brief  Sets pion velocity range
    /// \param  vMin  Minimal velocity [cm/ns]
    /// \param  vMax  Maximal velocity [cm/ns]
    void SetPionVelocityRange(double vMin, double vMax)
    {
      fMinBetaPion = vMin / kSpeedOfLight;
      fMaxBetaPion = vMax / kSpeedOfLight;
    }

    /// \brief  Sets proton velocity range
    /// \param  vMin  Minimal velocity [cm/ns]
    /// \param  vMax  Maximal velocity [cm/ns]
    void SetProtonVelocityRange(double vMin, double vMax)
    {
      fMinBetaProton = vMin / kSpeedOfLight;
      fMaxBetaProton = vMax / kSpeedOfLight;
    }

    /// \brief  Sets the assigned PDG for primary particles
    /// \param  pdg  PDG code of the particle
    void SetPrimaryAssignedPdg(int pdg) { fPrimaryAssignedPdg = pdg; }

    /// \brief  Assignes an uncertainty to the momentum measurement
    /// \param  uncertainty  Relative uncertainty ( = sqrt(var(q/p)) / (q/p))
    void SetQpAssignedUncertainty(double uncertainty) { fQpAssignedUncertainty = uncertainty; }

    /// \brief  Sets an offset to t0
    /// \param  offset  An offset [ns]
    void SetTzeroOffset(double offset) { fTzeroOffset = offset; }

    //* Specific parameters for the KFParticleFinder
    /// \brief  Sets cut on the distance to the primary vertex from the decay vertex
    /// \param  cut  Cut value [cm]
    void SetLCut(float cut) { GetKFParticleFinder()->SetLCut(cut); }

    /// \brief  Sets cut on \f$\chi^2_{prim}\f$ of each track for 2-daughter decays
    /// \param  cut  Cut value
    void SetChiPrimaryCut2D(float cut) { GetKFParticleFinder()->SetChiPrimaryCut2D(cut); }

    /// \brief  Sets cut on \f$\chi^2_{geo}\f$ for 2-daughter decays
    /// \param  cut  Cut value
    void SetChi2Cut2D(float cut) { GetKFParticleFinder()->SetChi2Cut2D(cut); }

    /// \brief  Sets cut on \f$l/\Delta l\f$ for 2-daughter decays
    /// \param  cut  Cut value
    void SetLdLCut2D(float cut) { GetKFParticleFinder()->SetLdLCut2D(cut); }

   private:
    /// \brief  Collects T0 values among the BMON hits
    /// \param  bmonHits  A span of BMON hits
    /// \return A vector of T0-s
    ///
    /// If multiple T0-s are found, the routine will run multiple times, until V0-candidates are found
    std::vector<double> CollectT0(gsl::span<const bmon::Hit> bmonHits) const;

    /// \brief  Collects a vector of DCA
    /// \param  recoEvent  Instance of a reconstructed event
    /// \return A vector of DCAs to origin
    std::vector<double> CollectDca(const RecoResults& recoEvent) const;

    /// \brief  Estimate DCA of a track to origin
    /// \param  fst  first  STS hit
    /// \param  snd  second STS hit
    /// \return dca [cm]
    double EstimateDca(const sts::Hit& fst, const sts::Hit& snd) const;

    //* Physical constants
    static constexpr double kPionMass{0.13957039};      ///< Pion mass [GeV/c2]
    static constexpr double kProtonMass{0.938272088};   ///< Proton mass [GeV/c2]
    static constexpr double kSpeedOfLight{29.9792458};  ///< Speed of light [cm/ns]


    V0FinderMonitorData_t fEventMonitor;  ///< Monitor data instance (per event)

    //* Different run-time cuts and flags (TODO: define in a config)
    double fTzeroOffset{0.};             ///< Offset for T0
    double fMinPionDca{1.5};             ///< Minimum DCA to PV for pions
    double fMinProtonDca{0.5};           ///< Minimum DCA to PV for protons
    double fQpAssignedUncertainty{0.1};  ///< Assigned relative uncertainty for q/p estimation
    int fPrimaryAssignedPdg{321};        ///< Assigned PDG hypothesis for primary particles
    double fMinBetaProton{0.};           ///< Minimal proton velocity (beta) [c]
    double fMaxBetaProton{1.};           ///< Maximal proton velocity (beta) [c]
    double fMinBetaPion{0.};             ///< Minimal proton velocity (beta) [c]
    double fMaxBetaPion{1.};             ///< Maximal proton velocity (beta) [c]


    //* Run-time variables, provided by framework
    int fBmonPartitionIndex{-1};                ///< Index of selected partition in BMON hit vector
    std::array<double, 3> fOrigin{0., 0., 0.};  ///< Coordinates of origin [cm]

    //* Auxilary variables

    /// \brief An instance of the topology reconstructor
    std::unique_ptr<KFParticleTopoReconstructor> fpTopoReconstructor{std::make_unique<KFParticleTopoReconstructor>()};
  };
}  // namespace cbm::algo::kfp
